from typing import Any

from markupsafe import Markup

from markupy import iter_node, render_node
from markupy.tag import Div, Tr


def assert_markup(result: Any, expected: str) -> None:
    assert isinstance(result, Markup)
    assert result == expected


class Test_render_node:
    def test_element(self) -> None:
        result = render_node(Div["a"])
        assert_markup(result, "<div>a</div>")

    def test_multiple_element(self) -> None:
        result = render_node(Tr["a"], Tr["b"])
        assert_markup(result, "<tr>a</tr><tr>b</tr>")

    def test_list(self) -> None:
        result = render_node([Tr["a"], Tr["b"]])
        assert_markup(result, "<tr>a</tr><tr>b</tr>")

    def test_none(self) -> None:
        result = render_node(None)
        assert_markup(result, "")

    def test_string(self) -> None:
        result = render_node("hello!")
        assert_markup(result, "hello!")


class Test_iter_node:
    def test_element(self) -> None:
        result = list(iter_node(Div["a"]))

        # Ensure we get str back, not markup.
        assert type(result[0]) is str
        assert result == ["<div>", "a", "</div>"]

    def test_list(self) -> None:
        result = list(iter_node([Tr["a"], Tr["b"]]))
        assert result == ["<tr>", "a", "</tr>", "<tr>", "b", "</tr>"]

    def test_none(self) -> None:
        result = list(iter_node(None))
        assert result == []

    def test_string(self) -> None:
        result = list(iter_node("yo!"))
        assert result == ["yo!"]



import keyword
import re
from html.parser import HTMLParser
from typing import Iterator

from markupsafe import escape

from markupy import tag

from .attribute import is_boolean_attribute
from .element import VoidElement

_void_elements: set[str] = {
    element.name
    for element in map(lambda x: getattr(tag, x), tag.__all__)
    if isinstance(element, VoidElement)
}


def _is_void_element(name: str) -> bool:
    return name in _void_elements


def _format_attribute_key(key: str) -> str:
    if keyword.iskeyword(key):
        return f"{key}_"
    if key.startswith("@"):
        key = f"_{key[1:]}"
    parts = key.split("-")
    key = f"{parts[0]}{''.join(map(lambda x: x.capitalize(), parts[1:]))}"
    key = key.replace(":", "__")
    key = key.replace(".", "_")
    return key


def _quote(value: str) -> str:
    quote = '"' if '"' not in value else "'" if "'" not in value else '"""'
    return f"{quote}{value}{quote}"


def _format_attribute_value(value: str | None) -> str:
    if value is None:
        return "True"
    return _quote(value)


def _format_attrs_dict(attrs: dict[str, str | None]) -> str:
    return "{" + ",".join(f'"{key}":{value}' for key, value in attrs.items()) + "}"


# Process attrs as they are received from the html parser
def _format_attrs(
    attrs: list[tuple[str, str | None]], *, use_selector: bool
) -> str | None:
    if not attrs:
        return None

    arguments: list[str] = []

    selector: str = ""
    attrs_kwargs: list[str] = list()
    attrs_dict: dict[str, str | None] = dict()

    for key, value in attrs:
        if is_boolean_attribute(key):
            value = None
        elif not value:
            continue
        elif key == "id" and use_selector and "{" not in value:
            selector = f"#{value}{selector}"
            continue
        elif key == "class" and use_selector and "{" not in value:
            selector = f"{selector}.{'.'.join(value.split())}"
            continue

        py_key = _format_attribute_key(key)
        if py_key.isidentifier():
            attrs_kwargs.append(f"{py_key}={_format_attribute_value(value)}")
        else:
            attrs_dict[key] = _format_attribute_value(value)

    if selector:
        arguments.append(f'"{selector}"')
    if attrs_kwargs:
        arguments += attrs_kwargs
    if attrs_dict:
        arguments.append(_format_attrs_dict(attrs_dict))

    if len(arguments) > 0:
        return f"({','.join(arguments)})"

    return None


class Stack:
    def __init__(self) -> None:
        self._list: list[str] = []

    def __len__(self) -> int:
        return len(self._list)

    def push(self, item: str) -> None:
        if not item:
            raise ValueError("Can't push None or empty string into stack")
        self._list.append(item)

    def peek(self) -> str | None:
        if len(self) > 0:
            return self._list[-1]
        return None

    def pop(self) -> str | None:
        if len(self) > 0:
            return self._list.pop()
        return None

    def __iter__(self) -> Iterator[str]:
        yield from self._list


class MarkupyParser(HTMLParser):
    def __init__(self, *, use_selector: bool, use_import_tag: bool) -> None:
        self.count_top_level: int = 0
        self.code_stack: Stack = Stack()
        self.unclosed_stack: Stack = Stack()
        self.imports: set[str] = set()
        self.use_import_tag: bool = use_import_tag
        self.use_selector: bool = use_selector
        super().__init__()

    def handle_starttag(self, tag: str, attrs: list[tuple[str, str | None]]) -> None:
        # print("Encountered a start tag:", tag, attrs)
        if len(self.unclosed_stack) == 0:
            self.count_top_level += 1
        if not _is_void_element(tag):
            self.unclosed_stack.push(tag)

        markupy_tag = "".join(map(lambda x: x.capitalize(), tag.split("-")))
        if self.use_import_tag:
            markupy_tag = f"tag.{markupy_tag}"

        self.imports.add(markupy_tag)
        self.code_stack.push(markupy_tag)
        if attributes_str := _format_attrs(attrs, use_selector=self.use_selector):
            self.code_stack.push(attributes_str)
        if _is_void_element(tag):
            self.code_stack.push(",")
        else:
            self.code_stack.push("[")

    def handle_endtag(self, tag: str) -> None:
        # print("Encountered an end tag :", tag)
        if not _is_void_element(tag):
            last_open_tag = self.unclosed_stack.pop()
            if last_open_tag is None:
                raise ValueError(f"Unexpected closing tag `</{tag}>`")
            elif tag != "endblock" and tag != last_open_tag:
                raise ValueError(
                    f"Invalid closing tag `</{tag}>`, expected `</{last_open_tag}>`"
                )
            elif tag == "endblock" and not last_open_tag.startswith("block-"):
                raise ValueError(
                    f"Invalid template `endblock`, expected `</{last_open_tag}>`"
                )

        if self.code_stack.peek() == ",":
            self.code_stack.pop()
        if self.code_stack.peek() == "[":
            self.code_stack.pop()
        elif not _is_void_element(tag):
            self.code_stack.push("]")
        self.code_stack.push(",")

    def handle_data(self, data: str) -> None:
        # print("Encountered some data  :", data)
        for line in data.splitlines():
            # Strip newlines, leading/traing spaces and redundant spaces
            line = " ".join(line.split())
            if not line:
                continue
            if len(self.unclosed_stack) == 0:
                self.count_top_level += 1
            self.code_stack.push(_quote(line))
            self.code_stack.push(",")

    def output_imports(self) -> str:
        if self.imports:
            if self.use_import_tag:
                return "from markupy import tag\n"
            else:
                return f"from markupy.tag import {','.join(sorted(self.imports))}\n"
        return ""

    def output_code(self) -> str:
        code = "".join(self.code_stack).strip(",")
        if self.count_top_level > 1:
            return f"[{code}]"
        return code


def _template_process(html: str) -> str:
    # Replace opening `block`
    html = re.sub(
        r"{%[+-]?\s+block\s+([a-zA-Z_]+)\s+[+-]?%}",
        lambda match: f"<block-{match.group(1).lower().replace('_','-')}>",
        html,
    )
    # Replace closing `block`
    html = re.sub(
        r"{%[+-]?\s+endblock(?:\s+[a-zA-Z_]+)?\s+[+-]?%}",
        # we can use whatever closing tag name we want here as it'll end up being replace with a closing bracket
        "</endblock>",
        html,
    )
    # Escape template contents
    html = re.sub(
        r"{{.*?}}|{%.*?%}|{#.*?#}",
        lambda match: escape(match.group(0)),
        html,
    )
    return html


def to_markupy(
    html: str, *, use_selector: bool = True, use_import_tag: bool = False
) -> str:
    parser = MarkupyParser(use_selector=use_selector, use_import_tag=use_import_tag)
    parser.feed(_template_process(html))
    parser.close()
    if tag := parser.unclosed_stack.pop():
        raise ValueError(f"Opening tag `<{tag}>` was not closed")
    if code := parser.output_code():
        return f"{parser.output_imports()}{code}"
    return ""



from markupy import Node, View, tag


class ComponentElement(View):
    def render(self) -> Node:
        return tag.Div


class ComponentElementMultiple(View):
    def render(self) -> Node:
        return tag.Div, tag.Img


class ComponentInComponent(View):
    def render(self) -> Node:
        return tag.Input, ComponentElement()


def test_component_element() -> None:
    assert str(ComponentElement()) == """<div></div>"""


def test_component_element_multiple() -> None:
    assert str(ComponentElementMultiple()) == """<div></div><img>"""


def test_component_in_component() -> None:
    assert str(ComponentInComponent()) == """<input><div></div>"""



import pytest

from markupy._private.html import to_markupy


def test_nested_void() -> None:
    html = """<div><hr></div>"""
    py = """from markupy.tag import Div,Hr\nDiv[Hr]"""
    assert to_markupy(html) == py


def test_empty_element() -> None:
    html = """<div></div>"""
    py = """from markupy.tag import Div\nDiv"""
    assert to_markupy(html) == py


def test_strip() -> None:
    html = """<div>\n</div><div>  \n  </div>"""
    py = """from markupy.tag import Div\n[Div,Div]"""
    assert to_markupy(html) == py


def test_doctype() -> None:
    html = """<!doctype html>"""
    py = ""
    assert to_markupy(html) == py


def test_selector() -> None:
    html = """<img id=test class="portait image">"""
    py = """from markupy.tag import Img\nImg("#test.portait.image")"""
    assert to_markupy(html) == py


def test_kwargs() -> None:
    html = """<label for="myfield" style="display:none" http-equiv="x">Hello</label>"""
    py = """from markupy.tag import Label\nLabel(for_="myfield",style="display:none",httpEquiv="x")["Hello"]"""
    assert to_markupy(html) == py


def test_empty_kwargs() -> None:
    html = """<input disabled="" style="">"""
    py = """from markupy.tag import Input\nInput(disabled=True)"""
    assert to_markupy(html) == py


def test_invalid_html_unclosed() -> None:
    html = """<div>"""
    with pytest.raises(ValueError):
        to_markupy(html)


def test_invalid_html_toomany_closed() -> None:
    html = """<div></div></div>"""
    with pytest.raises(ValueError):
        to_markupy(html)


def test_invalid_html_not_matching() -> None:
    html = """<div></pre>"""
    with pytest.raises(ValueError):
        to_markupy(html)


def test_to_markupy() -> None:
    html = """<html><Head><TITLE>Test</title></head><body class=''><h1 id='myid' burger&fries='good' class='title header'>Parse me! <!--My comment--></h1><hr><input class='my-input' disabled value='0' @click.outside.500ms='test' data-test='other' data-url-valid='coucou'><sl-button hx-on:htmx:config-request='attri'>Click!</sl-button></body></html>"""
    py = """from markupy.tag import Body,H1,Head,Hr,Html,Input,SlButton,Title\nHtml[Head[Title["Test"]],Body[H1("#myid.title.header",{"burger&fries":"good"})["Parse me!"],Hr,Input(".my-input",disabled=True,value="0",_click_outside_500ms="test",dataTest="other",dataUrlValid="coucou"),SlButton(hxOn__htmx__configRequest="attri")["Click!"]]]"""
    assert to_markupy(html) == py


def test_escape() -> None:
    html = """<a href="{{ url_for(".index") }}">Hello</a>"""
    py = """from markupy.tag import A\nA(href='{{ url_for(".index") }}')["Hello"]"""
    assert to_markupy(html) == py


def test_jinja() -> None:
    html = """
    <body>
    <h1>{{ heading }}</h1>
    <p>Welcome to our cooking site, {{ user.name }}!</p>

    <h2>Recipe of the Day: {{ recipe.name }}</h2>
    <p>{{ recipe.description }}</p>

    <h3>Instructions:</h3>
    <ol>
        {% for step in recipe.steps %}
        <li>{{ step }}</li>
        {% endfor %}
    </ol>
    </body>
    """
    py = """from markupy.tag import Body,H1,H2,H3,Li,Ol,P\nBody[H1["{{ heading }}"],P["Welcome to our cooking site, {{ user.name }}!"],H2["Recipe of the Day: {{ recipe.name }}"],P["{{ recipe.description }}"],H3["Instructions:"],Ol["{% for step in recipe.steps %}",Li["{{ step }}"],"{% endfor %}"]]"""
    assert to_markupy(html) == py


def test_self_closing() -> None:
    html = """<input type="checkbox" />"""
    py = """from markupy.tag import Input\nInput(type="checkbox")"""
    assert to_markupy(html) == py


def test_use_import_tag() -> None:
    html = """<div>hello</div>"""
    py = """from markupy import tag\ntag.Div["hello"]"""
    assert to_markupy(html, use_import_tag=True) == py


def test_use_selector() -> None:
    html = """<div id="myid" class="cls1 cls2">hello</div>"""
    py = """from markupy.tag import Div\nDiv(id="myid",class_="cls1 cls2")["hello"]"""
    assert to_markupy(html, use_selector=False) == py


def test_jinja_block() -> None:
    html = """
    {% block my_name %}
        <div></div>
    {% endblock %}
    """
    py = """from markupy.tag import BlockMyName,Div\nBlockMyName[Div]"""
    assert to_markupy(html) == py



from ._private.element import Element
from ._private.html import to_markupy as html2markupy
from ._private.view import Node, View, iter_node, render_node

__all__ = ["Element", "Node", "View", "html2markupy", "iter_node", "render_node"]
__version__ = "0.9.4"



from collections.abc import Callable, Generator, Iterable, Iterator
from typing import TypeAlias, final

from markupsafe import Markup, escape

Node: TypeAlias = None | bool | str | int | Iterable["Node"] | Callable[[], "Node"]


def _is_empty_node(node: Node) -> bool:
    return node is None or node is True or node is False or node == ""


def validate_node(node: Node) -> bool:
    if _is_empty_node(node):
        return False

    valid_nodes = str | int | View | Callable | Generator  # type: ignore[type-arg]
    invalid_nodes = bytes | bytearray | memoryview
    if isinstance(node, valid_nodes):
        return True
    elif isinstance(node, Iterable) and not isinstance(node, invalid_nodes):
        # Must return True if any child is valid
        # Must return False if all child not valid
        # Must loop over all items to raise exception any invalid child
        result = False
        for child in node:
            if validate_node(child):
                result = True
        return result
    else:
        raise TypeError(f"{node!r} is not a valid child element")


def iter_node(node: Node) -> Iterator[str]:
    if _is_empty_node(node):
        return
    while not isinstance(node, View) and callable(node):
        node = node()
    if isinstance(node, View):
        yield from node
    elif isinstance(node, int):
        yield str(node)
    elif isinstance(node, str):
        yield str(escape(node))
    elif isinstance(node, Iterable):
        for child in node:
            yield from iter_node(child)
    else:
        raise TypeError(f"{node!r} is not a valid child element")


def render_node(*node: Node) -> Markup:
    return Markup("".join(iter_node(node)))


class View(Iterable[str]):
    def render(self) -> Node:
        return None

    def __iter__(self) -> Iterator[str]:
        yield from iter_node(self.render())

    @final
    def __str__(self) -> str:
        return Markup("".join(self))
